''' Creating or importing backgrounds '''
from typing import Dict
import numpy as np
from inmotifin.utils.paramsdata.backgroundparams import BackgroundParams
from inmotifin.modules.data.background import Backgrounds
from inmotifin.modules.prepare.shuffler import Shuffler
from inmotifin.utils.fileops.reader import Reader
from inmotifin.utils.fileops.writer import Writer


class Backgrounder:
    ''' Class to generate or read background sequences
    Class parameters
    ----------------
    backgrounds: Backgrounds
        Data class for backgrounds
    '''

    def __init__(
            self,
            params: BackgroundParams,
            reader: Reader,
            writer: Writer,
            rng: np.random.Generator) -> None:
        ''' Constructor
        Parameters
        ----------
        params: BackgroundParams
            Dataclass storing alphabet, sequence length, sequence number, \
            alphabet_prior, background_files, to_shuffle \
            and number_of_shuffle
        reader: Reader
            File reader class to read in sequences if necessary
        writer: Writer
            instance of the writer class
        rng: np.random.Generator
            Random generator for sampling letters
        '''
        self.params = params
        self.reader = reader
        self.writer = writer
        self.rng = rng
        self.title = writer.get_title()
        self.backgrounds = None
        self.shuffler = Shuffler(
                number_of_shuffle=self.params.number_of_shuffle,
                rng=self.rng)

    def get_backgrounds(self) -> Backgrounds:
        ''' Getter for simulated backgrounds '''
        return self.backgrounds

    def get_backgrounds_seq(self) -> Backgrounds:
        ''' Getter for simulated backgrounds '''
        return self.backgrounds.backgrounds

    def create_backgrounds(self) -> None:
        ''' Controller function to read backgrounds or simulate if \
            no file available '''
        if self.params.background_files is not None:
            self.read_backgrounds()
        else:
            self.simulate_backgrounds()
            self.writer.dict_to_fasta(
                seq_dict=self.backgrounds.backgrounds,
                filename="simulated_backgrounds")

    def simulate_backgrounds(self) -> None:
        '''Generates a dictionary of random sequences with ids
        '''
        backgrounds = {}
        for seq_num in range(self.params.number_of_backgrounds):
            seq_id = self.title + "_seq_" + str(seq_num)
            alphabet_to_list = sorted(self.params.alphabet)
            random_list = [
                self.rng.choice(
                    alphabet_to_list,
                    p=self.params.alphabet_prior,
                    replace=True)[0]
                for _
                in range(self.params.length_of_backgrounds)]
            random_sequence = "".join(random_list)
            backgrounds[seq_id] = random_sequence
        if self.params.shuffle.lower() != "none":
            backgrounds = self.shuffle_backgrounds(backgrounds)

        self.backgrounds = Backgrounds(backgrounds=backgrounds)

    def read_backgrounds(self) -> None:
        ''' Generates a dictionary of sequences with ids '''
        backgrounds = self.reader.read_fasta(
            fasta_files=self.params.background_files)
        if self.params.shuffle.lower() != "none":
            backgrounds = self.shuffle_backgrounds(backgrounds)

        self.backgrounds = Backgrounds(backgrounds=backgrounds)

    def shuffle_backgrounds(
            self,
            backgrounds: Dict[str, str]) -> Dict[str, str]:
        ''' Shuffle available backgrounds thus generate new ones '''
        if self.params.shuffle.lower() == "random_nucl_addon":
            shuffled = self.shuffler.shuffle_seq_random_nucleotide(
                backgrounds=backgrounds)
            backgrounds.update(shuffled)
        elif self.params.shuffle.lower() == "random_nucl_only":
            backgrounds = self.shuffler.shuffle_seq_random_nucleotide(
                backgrounds=backgrounds)
        else:
            raise NotImplementedError
        return backgrounds
