''' Generate motif instances from motifs '''
from typing import List, Dict
import numpy as np
from inmotifin.utils import sequtils
from inmotifin.utils.baseutils import dict_of_list_to_dict
from inmotifin.modules.data.motif import Motifs


class MotifInstancer:
    ''' Class to take in motifs and generate \
    the required number of instances '''

    def __init__(
            self,
            motifs: Motifs,
            rng: np.random.Generator) -> None:
        ''' Constructor for motif instance generator
        Parameters
        ----------
        motifs: Motifs
            Data class for motifs with names (key) and PPM
        rng: np.random.Generator
            Random generator for multinomial instance sampling
        '''
        self.motifs = motifs
        self.rng = rng
        self.instances = {}

    def get_motif_instances(self, motif_index: int) -> List[str]:
        ''' Getter for the existing instances for a specific motif '''
        return self.instances[motif_index]

    def get_one_new_instance(
            self,
            motif_index: str) -> str:
        ''' Generate exactly one new motif instance '''
        motif = self.motifs.motifs[motif_index]
        l_motif = len(motif)
        instance_onehot = []
        for letter in range(l_motif):
            instance_onehot.append(
                self.rng.multinomial(1, motif[letter]))
        instance_str = sequtils.onehot_to_str(
            alphabet=self.motifs.alphabet,
            motif_onehot=instance_onehot)
        return instance_str

    def sample_instances(self, motif_idx_list: List[str]) -> Dict[str, str]:
        ''' Accessor function for creating new instances
        Parameters
        ----------
        motif_idx_list: List[str]
            List of motif IDs

        Return
        ------
        instances_unique_id: Dict[str, str]
            Dictionary of motif instances with unique ids and single elements
        '''
        instances = {}
        for motif_idx in motif_idx_list:
            instance_str = self.get_one_new_instance(
                motif_index=motif_idx)
            if motif_idx not in instances:
                instances[motif_idx] = [instance_str]
            else:
                instances[motif_idx].append(instance_str)
        self.update_instances_storage(new_instances=instances)
        instances_unique_id = dict_of_list_to_dict(indict=instances)
        return instances_unique_id

    def update_instances_storage(
            self,
            new_instances: Dict[str, List[str]]) -> None:
        ''' Add all new instances to the existing ones
        Parameters
        ----------
        new_instances: Dict[str, List[str]]
            Newly generated instances
        '''
        for motif_idx, inst in new_instances.items():
            if motif_idx not in self.instances:
                self.instances[motif_idx] = inst
            else:
                self.instances[motif_idx] += inst

    def orient_motifs(
            self,
            current_instances: Dict[str, str],
            orientations: List[int]) -> None:
        ''' Reverse complementing instances as necessary
        Parameters
        ----------
        instances_unique_id: Dict[str, str]
            Dictionary of motif instances with unique ids and single elements
        orientations: List[int]
            Mask for instances. List of 0s and 1s, where 0 means keeping the \
            orientation, 1 means reverse complementing the motif instance.
        '''
        oriented_instances = {}
        orient_idx = 0
        for inst_idx, instance in current_instances.items():
            if orientations[orient_idx] == 1:
                oriented_instances[inst_idx] = \
                    sequtils.create_reverse_complement(
                        alphabet=self.motifs.alphabet_revcomp_pairs,
                        motif_instance=instance)
            else:
                oriented_instances[inst_idx] = instance
            orient_idx += 1
        return oriented_instances
