""" Controller class to organize the preparation and sampling """
import os
import numpy as np
from inmotifin.utils.paramsdata.basicparams import BasicParams
from inmotifin.utils.paramsdata.motifparams import MotifParams
from inmotifin.utils.paramsdata.dimerparams import DimerParams
from inmotifin.utils.paramsdata.topicparams import TopicParams
from inmotifin.utils.paramsdata.freqparams import FreqParams
from inmotifin.utils.paramsdata.backgroundparams import BackgroundParams
from inmotifin.utils.paramsdata.samplingparams import SamplingParams
from inmotifin.utils.paramsdata.positionparams import PositionParams
from inmotifin.utils.fileops.reader import Reader
from inmotifin.utils.fileops.writer import Writer
from inmotifin.modules.prepare.setupper import Setupper
from inmotifin.modules.prepare.dimerer import Dimerer
from inmotifin.modules.prepare.motifer import Motifer
from inmotifin.modules.prepare.backgrounder import Backgrounder
from inmotifin.modules.sample.sampler import Sampler
from inmotifin.modules.sample.inminscheme import InMOTIFinScheme


class Controller:
    """ Organizer of preparation and sampling """

    def __init__(self, basic_params: BasicParams) -> None:
        """ Constructor """
        self.writer = Writer(
            workdir=basic_params.workdir,
            title=basic_params.title)
        self.reader = Reader()
        self.data_for_simulation = {}
        self.rng = np.random.default_rng(basic_params.seed)

    def create_dimers(self, dimer_params: DimerParams) -> None:
        """ Option of creating dimers given \
            input motifs and rules """
        my_dimerer = Dimerer(
            dimerparams=dimer_params,
            reader=self.reader,
            writer=self.writer,
            rng=self.rng)
        my_dimerer.create_dimers()

    def create_motifs(self, motif_params: MotifParams) -> None:
        """ Option of creating motifs given input parameters"""
        motifer = Motifer(
            params=motif_params,
            rng=self.rng,
            reader=self.reader,
            writer=self.writer)
        motifer.create_motifs()

    def create_backgrounds(self, background_params: BackgroundParams) -> None:
        """ Option of creating backgrounds given input parameters"""
        backgrounder = Backgrounder(
            params=background_params,
            reader=self.reader,
            writer=self.writer,
            rng=self.rng)
        backgrounder.create_backgrounds()

    def setup_simulation(
            self,
            motif_params: MotifParams,
            background_params: BackgroundParams,
            topic_params: TopicParams,
            freq_params: FreqParams) -> None:
        """ Create data for sampling """
        setupper = Setupper(
            reader=self.reader,
            writer=self.writer,
            motif_params=motif_params,
            background_params=background_params,
            topic_params=topic_params,
            frequency_params=freq_params,
            rng=self.rng)
        self.data_for_simulation["motifs"] = setupper.create_motifs()
        self.data_for_simulation["backgrounds"] = setupper.create_backgrounds()
        self.data_for_simulation["topics"] = setupper.create_topics()
        self.data_for_simulation["frequencies"] = setupper.create_frequencies()

    def run_sampling(
            self,
            sampling_params: SamplingParams,
            positions_params: PositionParams) -> None:
        """ Run main simulation module and save output
        """
        dag_name = self.writer.get_outfolder()
        sampler = Sampler(
            sampling_params=sampling_params,
            position_params=positions_params,
            data_for_simulation=self.data_for_simulation,
            reader=self.reader,
            rng=self.rng)
        my_sim = InMOTIFinScheme(
            dag_name=os.path.join(
                dag_name,
                self.writer.title + "_dagsim_table"),
            sampler=sampler,
            number_of_motived_seq=sampling_params.number_of_motived_seq)
        my_graph, data = my_sim.run_sampling()
        unmotived_seq = sampler.get_backgrounds(
            num_backgrounds=sampling_params.number_of_unmotived_seq)
        if sampling_params.to_draw:
            my_graph.draw()
        self.writer.save_dagsim_data(
            dagsim_data=data,
            unmotived_seq=unmotived_seq)

    def run_inmotifin(
            self,
            motif_params: MotifParams,
            background_params: BackgroundParams,
            topic_params: TopicParams,
            freq_params: FreqParams,
            sampling_params: SamplingParams,
            positions_params: PositionParams) -> None:
        """ Prepare and sample """
        self.setup_simulation(
            motif_params=motif_params,
            background_params=background_params,
            topic_params=topic_params,
            freq_params=freq_params)
        self.run_sampling(
            sampling_params=sampling_params,
            positions_params=positions_params)
