"""Main module."""
import click

from inmotifin.utils.config_args import (
    read_config,
    assign_background_params,
    assign_basic_params,
    assign_motif_params,
    assign_topic_freq_params,
    assign_position_sampling_params)
from inmotifin.utils.paramsdata.dimerparams import DimerParams
from inmotifin.organizer.controller import Controller


@click.group()
def cli():
    ''' You are running inMOTIFin '''
    click.echo('You are running inMOTIFin')


@cli.command()
@click.option(
    "--title", default="sim", type=str, help="Title of the analysis")
@click.option(
    "--workdir", default=".", type=click.Path(),
    help="Folder of the analysis outputs. Defaults to current work directory.")
@click.option(
    '--config', default=None, type=click.Path(),
    help="Config file for the simulation with the parameters for \
        creating motifs")
@click.option(
    "--dirichlet_alpha",
    help="Alpha values for the Dirichlet distribution \
        from which motifs are sampled.")
@click.option(
    "--num_motifs", type=int,
    help="Number of motifs to create.")
@click.option(
    "--m_length_min", type=int,
    help="Length of the motif. If --len_motifs_max is also \
            specified, this is the lower boundary of length.")
@click.option(
    "--m_length_max", type=int,
    help="Maximum allowed length of the motif.")
@click.option(
    "--m_alphabet", type=str,
    help="String of letters in the motif alphabet, e.g. ACGT")
@click.option(
    '--seed', type=int,
    default=None,
    help="Seed value for reproducibility. Recommended (Optional)")
@click.pass_context
def motifs(
        ctx,
        title,
        workdir,
        config,
        dirichlet_alpha,
        num_motifs,
        m_length_min,
        m_length_max,
        m_alphabet,
        seed):
    ''' Creating motifs given information content, length, and alphabet '''
    click.echo(
        'Creating motifs given information content, length, and alphabet')
    ctx = read_config(ctx=ctx, config=config)
    ctx, basic_params = assign_basic_params(
        ctx=ctx, title=title, workdir=workdir, seed=seed)
    ctx.default_map['alphabet_pairs'] = None
    ctx.default_map['motif_files'] = None
    ctx.default_map['jaspar_db_version'] = None

    ctx, motif_params = assign_motif_params(
        ctx=ctx,
        dirichlet_alpha=dirichlet_alpha,
        num_motifs=num_motifs,
        m_length_min=m_length_min,
        m_length_max=m_length_max,
        m_alphabet=m_alphabet,
        alphabet_pairs=None,
        motif_files=None,
        jaspar_db_version=None)
    controller = Controller(basic_params=basic_params)
    controller.create_motifs(motif_params=motif_params)


@cli.command()
@click.option(
    "--title", default="sim", type=str, help="Title of the analysis")
@click.option(
    "--workdir", default=".", type=click.Path(),
    help="Folder of the analysis outputs. Defaults to current work directory.")
@click.option(
    '--config', default=None, type=click.Path(),
    help="Config file for the simulation with the parameters for \
        creating dimers")
@click.option(
    '--motif_files', type=click.Path(),
    help="Path to the motif files. Meme, jaspar and csv \
        (with jaspar motif IDs) formats are accepted.")
@click.option(
    "--jaspar_db_version", type=str,
    help="Version of JASPAR database to use when fetching \
        with JASPAR motif IDs.")
@click.option(
    "--dimerisation_rules", type=click.Path(),
    help="Path to the dimerisation rule tsv files. \
        It should have two tab separated columns: \
        list of motif IDs separated by comma, and \
        list of distances between them spearated by comma.")
@click.option(
    '--seed', type=int, default=None,
    help="Seed value for reproducibility. Recommended (Optional)")
@click.pass_context
def dimers(
        ctx,
        title,
        workdir,
        config,
        motif_files,
        jaspar_db_version,
        dimerisation_rules,
        seed):
    ''' Dimerizing motifs given motifs and distances'''
    click.echo('Dimerizing motifs given motifs and distances')
    ctx = read_config(ctx=ctx, config=config)
    ctx, basic_params = assign_basic_params(
        ctx=ctx, title=title, workdir=workdir, seed=seed)
    if motif_files is not None:
        mfiles = [fname.strip() for fname in motif_files.split(',')]
        ctx.default_map['motif_files'] = mfiles
    if jaspar_db_version is not None:
        ctx.default_map['jaspar_db_version'] = jaspar_db_version
    if dimerisation_rules is not None:
        ctx.default_map['dimerisation_rules'] = dimerisation_rules
    dimer_params = DimerParams(
        motif_files=ctx.default_map["motif_files"],
        jaspar_db_version=ctx.default_map["jaspar_db_version"],
        dimerisation_rule_path=ctx.default_map["dimerisation_rules"])
    controller = Controller(basic_params=basic_params)
    controller.create_dimers(dimer_params=dimer_params)


@cli.command()
@click.option(
    "--title", type=str, default="sim", help="Title of the analysis")
@click.option(
    "--workdir", default=".", type=click.Path(),
    help="Folder of the analysis outputs. Defaults to current work directory.")
@click.option(
    '--config', default=None, type=click.Path(),
    help="Config file for the simulation with the parameters for \
        creating random sequences")
@click.option(
    "--num_backgrounds", type=int,
    help="Number of background sequences to simulate.")
@click.option(
    "--b_length", type=int,
    help="Length of background sequences to simulate.")
@click.option(
    "--b_alphabet", type=str,
    help="String of letters in the background alphabet, e.g. ACGT")
@click.option(
    "--alphabet_prior", type=str,
    help="Comma separated probability of the letters in the background \
        alphabet.")
@click.option(
    '--seed', type=int, default=None,
    help="Seed value for reproducibility. Recommended (Optional)")
@click.pass_context
def random_sequences(
        ctx,
        title,
        workdir,
        config,
        num_backgrounds,
        b_length,
        b_alphabet,
        alphabet_prior,
        seed):
    ''' Creating random sequences given length and alphabet '''
    click.echo(
        'Creating random sequences given length and alphabet')
    ctx = read_config(ctx=ctx, config=config)
    ctx, basic_params = assign_basic_params(
        ctx=ctx, title=title, workdir=workdir, seed=seed)
    ctx, background_params = assign_background_params(
        ctx=ctx,
        num_backgrounds=num_backgrounds,
        b_alphabet=b_alphabet,
        alphabet_prior=alphabet_prior,
        b_length=b_length,
        background_files=None,
        shuffle=None,
        num_shuffle=None)
    controller = Controller(basic_params=basic_params)
    controller.create_backgrounds(background_params=background_params)


@cli.command()
@click.option(
    "--title", default="sim", type=str, help="Title of the analysis")
@click.option(
    "--workdir", default=".", type=click.Path(),
    help="Folder of the analysis outputs. Defaults to current work directory.")
@click.option(
    '--config', default=None, type=click.Path(),
    help="Config file for the simulation with the parameters for \
        creating motived sequences")
@click.option(
    "--dirichlet_alpha",
    help="Alpha values for the Dirichlet distribution \
        from which motifs are sampled.")
@click.option(
    "--num_motifs", type=int,
    help="Number of motifs to create.")
@click.option(
    "--m_length_min", type=int,
    help="Length of the motif. If --len_motifs_max is also \
            specified, this is the lower boundary of length.")
@click.option(
    "--m_length_max", type=int,
    help="Maximum allowed length of the motif. (Optional) ")
@click.option(
    "--m_alphabet", type=str,
    help="String of letters in the motif alphabet, e.g. ACGT")
@click.option(
    "--motif_files", type=click.Path(),
    help="Path to the motif files. Meme, jaspar and csv \
        (with jaspar motif IDs) formats are accepted.")
@click.option(
    "--jaspar_db_version", type=str,
    help="Version of JASPAR database to use when fetching \
        with JASPAR motif IDs.")
@click.option(
    "--alphabet_pairs", type=str,
    help="Dictionary of letter pairs for reverse complementing the \
        motif instance e.g. 'A:T,C:G,G:C,T:A'")
@click.option(
    "--num_backgrounds", type=int,
    help="Number of background sequences to simulate.")
@click.option(
    "--b_length", type=int,
    help="Length of background sequences to simulate.")
@click.option(
    "--b_alphabet", type=str,
    help="String of letters in the background alphabet, e.g. ACGT")
@click.option(
    "--alphabet_prior", type=str,
    help="Comma separated probability of the letters in the background \
        alphabet.")
@click.option(
    "--background_files", type=click.Path(),
    help="Path to the background files in fasta format.")
@click.option(
    "--shuffle", type=str,
    help="Type of shuffling done on the backgrounds. \
        None or random are supported.")
@click.option(
    "--num_shuffle", type=int,
    help="Number of shuffle of the backgrounds.\
        Used when shuffle is set as random.")
@click.option(
    "--num_topics", type=int,
    help="Number of topics into which motifs are assigned.")
@click.option(
    "--max_topic_size", type=int,
    help="Maximum allowed motifs in a topic.")
@click.option(
    "--topic_size_p", type=float,
    help="P paremeter for sampling the size of each topic from \
        a binomial distribution.")
@click.option(
    "--topic_motif_assignment_file", type=click.Path(),
    help="Path to the motif to topic asisgnment file in two column \
        tsv format.")
@click.option(
    "--topic_freq_type", type=str,
    help="The method of selecting topic background frequencies.\
        Values: uniform, random. Where uniform means each topic has an \
        equal chance to be selected. Random means each topic is assigned \
        a probability of being selected. The difference between a \
        frequent and rare topic is controlled by the --topic_freq_range \
        parameter.")
@click.option(
    "--topic_freq_range", type=int,
    help="The range of the expected differences between a frequent \
        and a rare topic.")
@click.option(
    "--motif_freq_type", type=str,
    help="The method of selecting motif background frequencies.\
        Values: uniform, random. Where uniform means each motif has an \
        equal chance to be selected. Random means each motif is assigned \
        a probability of being selected. The difference between a \
        frequent and rare motif is controlled by the --motif_freq_range \
        parameter.")
@click.option(
    "--motif_freq_range", type=int,
    help="The range of the expected differences between a frequent \
        and a rare motif")
@click.option(
    "--concentration_factor", type=float,
    help="The preference of motifs to share the same topic when \
        selected for insertion")
@click.option(
    "--topic_topic_type", type=str,
    help="The method of selecting topic-topic co-occurence \
        probabilities. Values: uniform, random. Where uniform \
        means any two topics are equally probable of co-occuring. \
        Random means topic pairs are assigned a random probability \
        of co-occurence.")
@click.option(
    "--topic_freq_file", type=click.Path(),
    help="Tsv file including the background frequencies for the \
        topics to be selected")
@click.option(
    "--motif_freq_file", type=click.Path(),
    help="Tsv file including the background frequencies for the \
        selection of motifs to be inserted")
@click.option(
    "--topic_topic_file", type=click.Path(),
    help="Tsv file including the topic-topic cooccurence \
        probability matrix")
@click.option(
    "--position_simulated", type=str,
    help="Whether the positions of the motifs should be simulated. \
        Accepted values: central, left_central, right_central, uniform, \
        gaussian. Central means the first motif is inserted into the \
        center of the background. Uniform means all position has equal \
        chance. Gaussian means following the probabilities of a Gaussian \
        distribution centered on the given position of the background \
        as per the position_means and position_variances parameters.")
@click.option(
    "--position_means", type=str,
    help="Comma separated mean values for the gaussian positioning option.")
@click.option(
    "--position_variances", type=str,
    help="Comma separated variance values for the gaussian positioning \
        option.")
@click.option(
    "--position_file", type=click.Path(),
    help="Tsv file specifying the position where the motifs should \
        be inserted at. The first column can be integers so that \
        positions are randomly assigned for motifs, or motif IDs.\
        Mutually exclusive with the position_simulated parameter.")
@click.option(
    "--num_motived_seq", type=int,
    help="Number of sequences with motifs in them to generate.")
@click.option(
    "--pc_unmotived", type=int,
    help="Percentage of sequences without motifs.")
@click.option(
    "--to_replace", default=True, is_flag=True,
    help="Whether to replace backgorund bases with motif instance \
        (alternative is to insert between existing bases).")
@click.option(
    "--orientation_prob", type=float,
    help="Probability of reversing a motif instance in the motived \
        sequence.")
@click.option(
    "--num_topics_per_seq", type=int,
    help="Number of topics to sample (with replacement) per sequence.")
@click.option(
    "--n_instances_per_sequence", type=int,
    help="Number of motif instances to be inserted per sequence.")
@click.option(
    "--n_instances_per_sequence_l", type=int,
    help="Lambda parameter of Poisson distribution for selecting the \
        number of motif instances to be inserted per sequence.")
@click.option(
    '--to_draw', default=False, is_flag=True,
    help="Whether to draw the DAG")
@click.option(
    '--seed', type=int,
    default=None,
    help="Seed value for reproducibility. Recommended (Optional)")
@click.pass_context
def motived_sequences(
        ctx,
        title,
        workdir,
        config,
        dirichlet_alpha,
        num_motifs,
        m_length_min,
        m_length_max,
        m_alphabet,
        motif_files,
        jaspar_db_version,
        alphabet_pairs,
        num_backgrounds,
        b_length,
        b_alphabet,
        alphabet_prior,
        background_files,
        shuffle,
        num_shuffle,
        num_topics,
        max_topic_size,
        topic_size_p,
        topic_motif_assignment_file,
        topic_freq_type,
        topic_freq_range,
        motif_freq_type,
        motif_freq_range,
        concentration_factor,
        topic_topic_type,
        topic_freq_file,
        motif_freq_file,
        topic_topic_file,
        position_simulated,
        position_means,
        position_variances,
        to_replace,
        position_file,
        num_motived_seq,
        pc_unmotived,
        orientation_prob,
        num_topics_per_seq,
        n_instances_per_sequence,
        n_instances_per_sequence_l,
        to_draw,
        seed):
    ''' Simulating sequences with inserted motif instances '''
    click.echo('Simulating sequences with inserted motif instances')
    ctx = read_config(ctx=ctx, config=config)
    ctx, basic_params = assign_basic_params(
        ctx=ctx, title=title, workdir=workdir, seed=seed)
    ctx, motif_params = assign_motif_params(
        ctx=ctx,
        dirichlet_alpha=dirichlet_alpha,
        num_motifs=num_motifs,
        m_length_min=m_length_min,
        m_length_max=m_length_max,
        m_alphabet=m_alphabet,
        alphabet_pairs=alphabet_pairs,
        motif_files=motif_files,
        jaspar_db_version=jaspar_db_version)
    ctx, background_params = assign_background_params(
        ctx=ctx,
        num_backgrounds=num_backgrounds,
        b_alphabet=b_alphabet,
        alphabet_prior=alphabet_prior,
        b_length=b_length,
        background_files=background_files,
        shuffle=shuffle,
        num_shuffle=num_shuffle)
    ctx, topic_params, freq_params = assign_topic_freq_params(
        ctx=ctx,
        num_topics=num_topics,
        max_topic_size=max_topic_size,
        topic_size_p=topic_size_p,
        topic_motif_assignment_file=topic_motif_assignment_file,
        topic_freq_type=topic_freq_type,
        topic_freq_range=topic_freq_range,
        motif_freq_type=motif_freq_type,
        motif_freq_range=motif_freq_range,
        concentration_factor=concentration_factor,
        topic_topic_type=topic_topic_type,
        topic_freq_file=topic_freq_file,
        motif_freq_file=motif_freq_file,
        topic_topic_file=topic_topic_file)
    ctx, positions_params, sampling_params = assign_position_sampling_params(
        ctx=ctx,
        position_simulated=position_simulated,
        position_means=position_means,
        position_variances=position_variances,
        position_file=position_file,
        num_motived_seq=num_motived_seq,
        pc_unmotived=pc_unmotived,
        to_replace=to_replace,
        orientation_prob=orientation_prob,
        num_topics_per_seq=num_topics_per_seq,
        n_instances_per_sequence=n_instances_per_sequence,
        n_instances_per_sequence_l=n_instances_per_sequence_l,
        to_draw=to_draw)
    controller = Controller(basic_params=basic_params)
    controller.run_inmotifin(
        motif_params=motif_params,
        background_params=background_params,
        topic_params=topic_params,
        freq_params=freq_params,
        sampling_params=sampling_params,
        positions_params=positions_params)
