""" Output class for methods writing in data from various formats """
from typing import List, Dict, Any
import os
import numpy as np
import pandas as pd


class Writer:
    """IO methods for saving simulated motifs, topics and backgrounds"""

    def __init__(self, workdir: str, title: str) -> None:
        ''' Constructor '''
        self.workdir = workdir
        self.title = title
        self.outfolder = os.path.join(self.workdir, self.title)
        self.setup_dirs()

    def get_title(self) -> str:
        ''' Getter for the analysis title '''
        return self.title

    def get_outfolder(self) -> str:
        ''' Getter for name of folder where output is saved '''
        return self.outfolder

    def setup_dirs(self) -> None:
        ''' Create workfolder and outfolder if does not exist yet '''
        os.makedirs(self.workdir, exist_ok=True)
        os.makedirs(self.outfolder, exist_ok=True)

    def list_to_file(
            self,
            data_list: List,
            filename: str,
            file_format: str = "txt") -> None:
        '''Export data in list format to a file
        One entry = one line
        Parameters
        ----------
        data_list:
            any type of data in a list format
        filename:
            output file name (without extension)
        '''
        filepath = os.path.join(
            self.outfolder,
            self.title + '_' + filename + '.' + file_format)
        with open(filepath, 'w+', encoding='utf-8') as out_file:
            for line in data_list:
                out_file.write(f"{line}\n")

    def pandas_to_tsv(self, dataframe: pd.DataFrame, filename: str) -> None:
        ''' Write pandas dataframe to file '''
        filepath = os.path.join(
            self.outfolder,
            self.title + '_' + filename + ".tsv")
        dataframe.to_csv(filepath, sep='\t')

    def dict_to_tsv(
            self,
            data_dict: Dict[Any, List[Any]], filename: str) -> None:
        '''Export data in dict format to a tsv file
        One entry = one line
        Parameters
        ----------
        data_dict Dict[Any, List[Any]]:
            any type of data in a dictionary format
        filename:
            output file name (without extension)
        '''
        datalist = list(data_dict.items())
        outputs = []
        datalines = [f"{k}\t{v}" for k, v in datalist]
        removal_dict = {
            ord('['): None,
            ord(']'): None,
            ord(' '): None,
            ord('\''): None}
        for values in datalines:
            outputs.append(f"{values.translate(removal_dict)}")
        self.list_to_file(
            data_list=outputs,
            filename=filename,
            file_format="tsv")

    def dict_to_fasta(
            self,
            seq_dict: Dict[str, str],
            filename: str) -> None:
        '''Export data in list format to a fasta file
        One entry = one line
        Parameters
        ----------
        data_dict Dict[str, str]:
            sequence info and actual sequence in a dictionary format
        filename:
            output file name (without extension)
        '''
        line_list = []
        for idx, seq in seq_dict.items():
            line_list.append(f">{idx}\n{seq}")
        self.list_to_file(
            data_list=line_list,
            filename=filename,
            file_format="fa")

    def motif_to_meme(
            self,
            motifs: Dict[str, np.ndarray],
            alphabet: str,
            file_prefix: str) -> None:
        """ Save motif position probability matrices in meme format

        Parameters
        ----------
        motif_id: str
            the ID of the simulated motif
        motifs: Dict[str, np.ndarray]
            the IDs and ppms of the simulated motifs
        alphabet: str
            Characters in the order of column assignment (eg ACGT)
        file_prefix: str
            name for the meme file without extension
        """
        outputs = []
        header = "MEME version 4\n\n" + \
            f"ALPHABET= {alphabet}"
        outputs.append(header)
        removal_dict = {ord('['): None, ord(']'): None}

        for motif_id in motifs:
            motif_header = f"\nMOTIF {motif_id}\n" + \
                f"letter-probability matrix: alength= {len(alphabet)} " + \
                f"w= {motifs[motif_id].shape[0]}"
            outputs.append(motif_header)

            formatted_values = np.apply_along_axis(
                np.array2string,
                axis=1,
                arr=motifs[motif_id],
                formatter={'float_kind': lambda x: f"{x:.6f}"}
            )
            for values in formatted_values:
                outputs.append(f"{values.translate(removal_dict)}")

        self.list_to_file(
            data_list=outputs,
            filename=file_prefix,
            file_format="meme")

    def motived_and_plain_to_fasta(
            self,
            data_for_fasta: Dict[int, Dict[str, Any]],
            unmotived_for_fasta: List[str]) -> None:
        """ Save fasta files with or without motived sequences.

        Parameters
        ----------
        data_for_fasta: Dict[int, Dict[str, Any]]
            Nested dictionary modified from the Dagsim output
        unmotived_seq: List[str]
            List of selected (not necessarily unique) background \
            IDs and sequences without inserted motif
        """
        filename = "final_sequences"
        output_dict = {}
        unique_id = -1
        for unique_id, info_dict in data_for_fasta.items():
            identifier = str(unique_id) + "_" + info_dict["background_id"]
            motived_seq = info_dict["motived_seq"]
            output_dict[identifier] = motived_seq
        for empty_sequence in unmotived_for_fasta:
            unique_id += 1
            identifier = str(unique_id) + "_" + empty_sequence.split(',')[0]
            output_dict[identifier] = empty_sequence.split(',')[1]

        self.dict_to_fasta(
            seq_dict=output_dict,
            filename=filename)

    def dagsim_data_to_dict_of_dict(
            self,
            dagsim_data: Dict[str, Any]) -> Dict[int, Dict[str, Any]]:
        """ Takes the dictionary generate by dagsim and transforms it

        Parameters
        ----------
        dagsim_data:
            dictionary from the Dagsim output

        Return
        ------
        transformed_data_dict: Dict[int, Dict[str, Any]]
            Dictionary of dictionary, where the outer key  are unique \
            identifiers to each simulated sequence and each dictionary \
            contains all the information to that sequence.
        """
        data_pd = pd.DataFrame(dagsim_data)
        transformed_data_dict = data_pd.to_dict(orient='index')

        return transformed_data_dict

    def data_to_bed(
            self,
            data_for_bed: Dict[int, Dict[str, Any]]) -> None:
        """ Save bed files with motif coordinates.

        Parameters
        ----------
        data_for_bed: Dict[int, Dict[str, Any]]
            Nested dictionary modified from the Dagsim output
        """
        filename = "inserted_instances"

        line_list = []
        for unique_id, info_dict in data_for_bed.items():

            num_instances = info_dict["num_instances"]
            positions = info_dict["positions"]
            orientations = info_dict["orientation"]
            names = info_dict["selected_motifs"]
            instances = info_dict["instances"]
            identifier = str(unique_id) + "_" + info_dict["background_id"]

            for instances_index in range(num_instances):
                orientation = \
                    '+' if orientations[instances_index] == 1 else '-'
                line_list.append(
                    f"{identifier}" +
                    f"\t{positions[instances_index][0]}" +
                    f"\t{positions[instances_index][1]}\t" +
                    f"{names[instances_index]}_" +
                    f"{instances[instances_index]}" +
                    "\t." +
                    f"\t{orientation}"
                )
        self.list_to_file(
            data_list=line_list,
            filename=filename,
            file_format="bed")

    def save_dagsim_data(
            self,
            dagsim_data: Dict[str, Any],
            unmotived_seq: Dict[str, str]) -> None:
        """ Save dagsim data and unmotived sequences to bed and fasta files
        Parameters
        ----------
        dagsim_data: Dict[str, Any]
            dictionary from the Dagsim output
        unmotived_seq: Dict[str, str]
            List of background sequences without inserted motif
        """
        transformed_data_dict = self.dagsim_data_to_dict_of_dict(dagsim_data)

        self.data_to_bed(data_for_bed=transformed_data_dict)
        self.motived_and_plain_to_fasta(
            data_for_fasta=transformed_data_dict,
            unmotived_for_fasta=unmotived_seq)
